// Easy SD Logger + VFD Display
// 2013.6.23 Eleken (rev_c1815@yahoo.co.jp)
// See also: http://eleken.futene.net/

// MCU : ATmega88
// main clock : 8.0 MHz
// Fuse High: 0b11010101 = 0xD5 (BOD = 2.9V)
// Fuse Low : 0b11100010 = 0xE0

/**********************************************
 Serial Data Format:
 
- 'm' + x : change receive mode {x = '0'-'2'}
 {0: ascii string mode, 1: ascii digit mode, 2: direct digit mode}
- 'b' + x : set brightness {x = 'A'(bright) - 'Z'(dark)}
- 'c' + x : set DP/Colon digits.
-In ascii string mode, successive ascii bytes (capital chars) are received while received bytes are from 0x20(' ') to 0x5F('_').
  After each bytes received, the display is updated and scrolled automatically.
-In ascii digit mode, 9-byte received data after mode change message is the ascii char (or DP/Colon at 9th digit) of each digits.
-In direct digit mode, 2*9-bytes received data after mode change message are segment data of each digits (MSB First).
- The default mode is ascii string mode.

***********************************************/

#include "main.h"
#include "YSwitchProc.h"
#include "msm62x42b.h"
#include "fatfs/pff.h"

/* Grobal Vars. */
sTime g_curTime;

// VFD Digids
volatile uint8_t g_digitData[10];
// index 0-7: digits 8(left) - 1(right), respectively.
// index 8  : (MSB) a_c_e_gh (LSB)
//            a: Dp7, c: col7, h: Dp3, e: col3, g: col5
// index 9  : blinking on/off

// Analog Pins
volatile uint16_t g_analogState[5]; // AIN0-4 + Switch

// Switch state
volatile uint8_t g_swState;

#define ISR_TIMER1_CALLED 1
#define ISR_PCINT1_CALLED 2
#define ISR_LOGGING_INT   4
#define ISR_USART_RX      8
#define ISR_CHECK_SWITCH  16
volatile uint8_t g_isrFlag;

enum tDisplayState{
	STATE_CLOCK1 = 0, // "AM 00:00:00"
	STATE_ADC1, // 
	STATE_ADC2, // 
	STATE_ADC3, // 
	STATE_ADC4, // 
	STATE_LOGGER, // "LOGGING"
	STATE_THRU, // "RX THRU "
	
	// Adjust time
	STATE_ADJUST_TIME_YEAR,
	STATE_ADJUST_TIME_MONTH,
	STATE_ADJUST_TIME_DAY,
	STATE_ADJUST_TIME_HOUR,
	STATE_ADJUST_TIME_MIN,
	
	// Logger setting
	STATE_LOGGER_SETTING_1,
//	STATE_LOGGING,
	
} g_curDisplayState;

typedef struct {
	uint8_t  logging;
	uint8_t  intervalIdx; // 1, 2, 10, 30, 60, 600
	uint16_t  intervalCnt; // count compare value
}sLoggerStatus;
#define INTERVAL_MAX 5

sLoggerStatus g_loggerStatus;


// EEPROM 
typedef struct {
	enum tDisplayState displayState;
	uint8_t brightness; // 0-31
	
	// sLoggerStatus
	uint8_t  interval; 
}sEEPSettings; // g_eep

volatile sEEPSettings g_eep;

/* EEPROM */
uint8_t eep_dummy __attribute__((section(".eeprom")));
uint8_t eep_area[sizeof(g_eep)] __attribute__((section(".eeprom")));

#if (ENABLE_EEP)
static void LoadEEP_sub(void){
	uint8_t i;
	uint8_t* p = (uint8_t*)(&g_eep);
	for(i=0; i<sizeof(g_eep); i++){
		uint8_t a = eeprom_read_byte(&(eep_area[i]));
		if(a != 0xFF) p[i] = a;
	}
}
static void SaveEEP_sub(void){
	uint8_t i;
	uint8_t* p = (uint8_t*)(&g_eep);
	for(i=0; i<sizeof(g_eep); i++){
		eeprom_write_byte(&(eep_area[i]), p[i]);
	}
}

static void LoadEEP(void){ // EEPROM -> RAM
	LoadEEP_sub();
	
	g_curDisplayState = g_eep.displayState;
	if(g_curDisplayState > STATE_THRU) g_curDisplayState = STATE_CLOCK1;
	if(g_eep.brightness > 31) g_eep.brightness = 6;
	if(g_eep.brightness < 2)  g_eep.brightness = 6;
	if(g_eep.interval > INTERVAL_MAX)  g_eep.interval = 0;
	g_loggerStatus.intervalIdx = g_eep.interval;
}
static void SaveEEP(void){ // RAM -> EEPROM 
	g_eep.displayState = g_curDisplayState;
	g_eep.interval = g_loggerStatus.intervalIdx;

	SaveEEP_sub();
}

#endif

/* Declare Function */
void FormatMsg(char* output, uint8_t length, const char* format, unsigned int x1);
void UpdateDisplayDigits(void);
void SendToDisplay(const uint8_t* digitData);
static void UART_Send(const uint8_t d);
#if (ENABLE_BRIGHTNESS)
static void UpdateBrightness(void);
#endif
void FormatLogFile(char* output);

static void UART_Send(const uint8_t d){
	while( (UCSR0A & 0b00100000) == 0); // wait until transmit complete
	UDR0 = d;
}


static void BCD_add(uint8_t* p){
	if(((*p) & 0x0F) == 9) (*p) += 7;
	else (*p) ++;
}
static void BCD_dec(uint8_t* p){
	if(((*p) & 0x0F) == 0) (*p) -= 7;
	else (*p) --;
}

static uint8_t BCD_sub(uint8_t a, uint8_t b){
	while(b&0x0F) {
		BCD_dec(&a);
		b--;
	}
	return a-b;
}
static void BCD_print(char* p, uint8_t v){
	p[0] = '0' + (v>>4);
	p[1] = '0' + (v&0x0F);
}
/*
static void utoa_usr(uint16_t v, char* p){
	uint8_t cnt = 0;
	uint8_t i = 0;
	uint16_t lref[] = {10000,1000,100,10,1};
	for(i=0; i<5; i++){
		uint16_t ref = lref[i];
		if(cnt || v>= ref){
			p[cnt] = '0';
			while(v>= ref){
				p[cnt] ++;
				v -=  ref;
			}
			cnt ++;
		}
	}
}*/


/* UART */
ISR(USART_RX_vect){
	if(g_curDisplayState == STATE_THRU){
		uint8_t d = UDR0;
		UART_Send(d);
	}
}

/* PCINT12 (PCIE1) Change */
ISR(PCINT1_vect){
	g_isrFlag |= ISR_PCINT1_CALLED;
	/*
	if(PINC & (1<<PC4)){
		// interrupt per 1sec.
		MSM62X42B_GetCurTime(&g_curTime);
		if(g_curDisplayState == STATE_CLOCK1){
			UpdateDisplayDigits();
			SendToDisplay((const uint8_t*)g_digitData);
		}
		
		// start 1-shot timer for colon blinking
		TCCR1B |= 0x05; // START TIMER1
		TCNT1   = 0;
		
	}
	*/
}

/* ADC conversion complete */
ISR(ADC_vect){

	uint8_t curChannel = ADMUX & 0x0F;
	if(curChannel <= 2){ // channel was below AIN2
		ADMUX = (ADMUX & 0xF0) | (curChannel+1);
		g_analogState[curChannel] = ADC;
		ADCSRA |= 0b01000000; // Start next conversion
	}else if(curChannel == 3){ // channel was AIN3
		ADMUX = (ADMUX & 0xF0) | 5;
		g_analogState[3] = ADC;
		ADCSRA |= 0b01000000; // Start next conversion
	}else{// channel was AIN5
		g_analogState[4] = ADC;
		
		// Update display
		static uint8_t cnt;
		if(++cnt == 5){
			if(g_curDisplayState >= STATE_ADC1 && g_curDisplayState <= STATE_ADC4){
				UpdateDisplayDigits();
				SendToDisplay((const uint8_t*)g_digitData);
			}
			cnt = 0;
		}
	}
	
}

/* Timer */

// timer0 : 10ms
ISR(TIMER0_COMPA_vect){
	static uint8_t cnt;
	
	if(++cnt == 10){
		g_isrFlag |= ISR_CHECK_SWITCH;
		/*
		// Check Switch State
		uint8_t sw = 0xFF;
		{
			uint8_t a = (g_analogState[4] >> 2);
			if(a < 0+32) sw = 0b11111110; // SW1: MENU
			else if(a < 64+32)  sw = 0b11111101; // SW2
			else if(a < 128+32) sw = 0b11111011; // SW3
			else if(a < 192+32) sw = 0b11110111; // SW4
		}
		
		GetSwitchState(sw, (uint8_t*)&g_swState);
		
		// Start acquiring next analog value
		ADMUX = (ADMUX & 0xF0) | 0;
		ADCSRA |= 0b01000000; // Start A-D conversion
		*/
		cnt = 0;
	}
}

// timer1 : 500ms from PCINT1
ISR(TIMER1_COMPA_vect){
	g_isrFlag |= ISR_TIMER1_CALLED;
	/*
	// STOP TIMER1
	TCCR1B &= ~0x07;
	if(g_curDisplayState == STATE_CLOCK1){
		// clear colon
		g_digitData[8] &= 0b11110001;
		SendToDisplay((const uint8_t*)g_digitData);
	}
	*/
}


// update g_digitData respect to the current state
void UpdateDisplayDigits(void){
	g_digitData[8] = 0;
	g_digitData[9] = 0;
	
	if(g_curDisplayState == STATE_CLOCK1){
		if(g_curTime.hour >= 0x12){
			uint8_t hp = BCD_sub(g_curTime.hour, 0x12);
			g_digitData[0] = 'P';
			g_digitData[2] = (hp&0xF0)? '0' + (hp>>4): ' ';
			g_digitData[3] = '0' + (hp&0x0F);
		}else{
			g_digitData[0] = 'A';
			g_digitData[2] = ((g_curTime.hour)&0xF0)? '0' + ((g_curTime.hour)>>4): ' ';
			g_digitData[3] = '0' + ((g_curTime.hour)&0x0F);
		}
		g_digitData[1] = 'M';
		g_digitData[4] = '0' + ((g_curTime.min)>>4);
		g_digitData[5] = '0' + ((g_curTime.min)&0x0F);
		g_digitData[6] = '0' + ((g_curTime.sec)>>4);
		g_digitData[7] = '0' + ((g_curTime.sec)&0x0F);
		g_digitData[8] = 0b00000011; // colon
	}
	else if(g_curDisplayState == STATE_ADC1){
		FormatMsg((char*)g_digitData, 8, "CH1 %d", g_analogState[0]);
	//	strcpy_P((char*)g_digitData, PSTR("CH1 "));
	//	utoa(g_analogState[0], (char*)&(g_digitData[4]), 10);
	}
	else if(g_curDisplayState == STATE_ADC2){
		FormatMsg((char*)g_digitData, 8, "CH2 %d", g_analogState[1]);
	}
	else if(g_curDisplayState == STATE_ADC3){
		FormatMsg((char*)g_digitData, 8, "CH3 %d", g_analogState[2]);
	}
	else if(g_curDisplayState == STATE_ADC4){
		FormatMsg((char*)g_digitData, 8, "CH4 %d", g_analogState[3]);
	}
	else if(g_curDisplayState == STATE_LOGGER){
		strcpy_P((char*)g_digitData, PSTR("LOGGING "));
	}
	else if(g_curDisplayState == STATE_THRU){
		strcpy_P((char*)g_digitData, PSTR("RX THRU "));
	}
	else if(g_curDisplayState == STATE_ADJUST_TIME_YEAR){
		strcpy_P((char*)g_digitData, PSTR("YEAR20"));
		g_digitData[6] = '0' + (g_curTime.year>>4);
		g_digitData[7] = '0' + (g_curTime.year&0x0F);
		g_digitData[9] = 0b00000011;
	}
	else if(g_curDisplayState == STATE_ADJUST_TIME_MONTH){
		strcpy_P((char*)g_digitData, PSTR("MONTH "));
		g_digitData[6] = '0' + (g_curTime.month>>4);
		g_digitData[7] = '0' + (g_curTime.month&0x0F);
		g_digitData[9] = 0b00000011;
	}
	else if(g_curDisplayState == STATE_ADJUST_TIME_DAY){
		strcpy_P((char*)g_digitData, PSTR("DATE  "));
		g_digitData[6] = '0' + (g_curTime.day>>4);
		g_digitData[7] = '0' + (g_curTime.day&0x0F);
		g_digitData[9] = 0b00000011;
	}
	else if(g_curDisplayState == STATE_ADJUST_TIME_HOUR){
		strcpy_P((char*)g_digitData, PSTR("HOUR  "));
		g_digitData[6] = '0' + (g_curTime.hour>>4);
		g_digitData[7] = '0' + (g_curTime.hour&0x0F);
		g_digitData[9] = 0b00000011;
	}
	else if(g_curDisplayState == STATE_ADJUST_TIME_MIN){
		strcpy_P((char*)g_digitData, PSTR("MIN   "));
		g_digitData[6] = '0' + (g_curTime.min>>4);
		g_digitData[7] = '0' + (g_curTime.min&0x0F);
		g_digitData[9] = 0b00000011;
	}
	
	else if(g_curDisplayState == STATE_LOGGER_SETTING_1){
		if(g_loggerStatus.intervalIdx == 0){
			strcpy_P((char*)g_digitData, PSTR("INT  1 S"));
		}
		else if(g_loggerStatus.intervalIdx == 1){
			strcpy_P((char*)g_digitData, PSTR("INT  2 S"));
		}
		else if(g_loggerStatus.intervalIdx == 2){
			strcpy_P((char*)g_digitData, PSTR("INT 10 S"));
		}
		else if(g_loggerStatus.intervalIdx == 3){
			strcpy_P((char*)g_digitData, PSTR("INT 30 S"));
		}
		else if(g_loggerStatus.intervalIdx == 4){
			strcpy_P((char*)g_digitData, PSTR("INT 60 S"));
		}
		else if(g_loggerStatus.intervalIdx == 5){
			strcpy_P((char*)g_digitData, PSTR("INT  5 M"));
		}
		g_digitData[8] = 0b00001000;
		g_digitData[9] = 0b00011111;
	}
}
// 
void SendToDisplay(const uint8_t* digitData){
	// Set to ascii digit mode
	UART_Send('m');
	UART_Send('1');
	// Send digit data (9-bytes)
	for(uint8_t i=0; i<10; i++){
		UART_Send(digitData[i]);
	}
}

void InitPort(void){
	
	// init clock
//	CLKPR = 0x80;
//	CLKPR = 0x00; // CLKPS = 0 (8MHz)
	
//	MCUSR = 0;
	
//	uint8_t i;
//	for(i=0; i<100; i++) _delay_ms(2);
	
	// init pins
	PORTB = 0b00010111;
	DDRB  = 0b11101111;
	
	PORTC = 0b00010000;
//	DDRC  = 0b00000000;
	
	PORTD = 0b00000011;
	DDRD  = 0b11111110;
}

// Initialize
// (To keep the program size small, initialization of the same register to default is comment-outed.)
uint8_t Init(void){
	
	cli();
	InitPort();
	
//	MSM62X42B_InitPort();
	uint8_t i;
	for(i=0;i<100;i++) _delay_ms(10);
	SBI(PORTD, RTC_CS1);
	
//	MSM62X42B_Reset(); // if battery-backup is diasbled...
	MSM62X42B_EnableOutput();
	MSM62X42B_GetCurTime(&g_curTime);
	
	// init digits
	
	// init SPI (clock:4.0MHz)
	SPCR = 0b01010000;// SPI mode 0
//	SPSR = 0b00000000;
	
	// init timer
	// timer0 (10ms = 8MHz/1024/78)
	TCCR0A = 0b00000010; // CTC
	TCCR0B = 0b00000101; // Clock -> clk/1024
	OCR0A  = 78;
	TIMSK0 = 0b00000010; // OCIE0A
//	TCNT0  = 0;
	
	// timer1 (500ms = 8MHz/1024/3906)
//	TCCR1A = 0b00000000; // CTC, TOP=OCR1A
//	TCCR1B = 0b00001101; // CK/1024
	TCCR1B = 0b00001000; // STOP TIMER1
	OCR1A  = 3906;
//	TCCR1C = 0;
//	TCNT1  = 0;
	TIMSK1 = 0b00000010; // OCIE1A
	
	
	// init USART
//	UBRR0H = 0;
	UBRR0L = 207; // 2400 bps at clk=8.0 MHz
//	UCSR0A = 0b00000000;
	UCSR0B = 0b10011000; // RXCIE, RXEN, TXEN
	UCSR0C = 0b00000110; // 8-bit, 1 Stop bit, asynchronous USART
	
	// init ADC
//	ADMUX  = 0b00000000; // MUX=0, VREF = AREF
	ADCSRA = 0b10001110; // CK/64, interrupt enable
//	ADCSRB = 0b00000000;
	DIDR0  = 0b00101111;
	
	g_analogState[4] = 1023;
	g_curDisplayState = STATE_CLOCK1;
	
	// init ext. interrupt (PCINT12)
	PCICR  = 0b00000010;
//	PCMSK2 = 0;
	PCMSK1 = 0b00010000;
//	PCMSK0 = 0;

	/*
	uint8_t buf[18];
	buf[0] = 0x51;
	buf[1] = 0x54;
	buf[2] = 0x40;
	buf[3] = 0xC0;
	buf[4] = 0x12;
	buf[5] = 0x20;
	buf[6] = 0b01000101;
	buf[7] = 0b01011000;
	buf[8] = 0b01010001;
	buf[9] = 0b01010011;
//	buf[6] = 0b01010011;
//	buf[7] = 0b11010001;
//	buf[8] = 0b01010011;
//	buf[9] = 0b11100000;
	buf[10] = 0x40;
	buf[11] = 0;
	buf[12] = 0x40;
	buf[13] = 0;
	buf[14] = 0x40;
	buf[15] = 0;
	buf[16] = 0;
	buf[17] = 0x00;
	
	UART_Send('m');
	UART_Send('2');
	// Send digit data (9-bytes)
	for(uint8_t i=0; i<18; i++){
		UART_Send(buf[i]);
	}
	while(1);
	*/
	
	sei();
	
	return 0;
	
}

// Example: FormatMsg(a, 10, "x = [%d]", 30) => a: "x = [30]  "
// buffer size must be at least (length + 1)
// This function is for debug use: not check output buffer length
void FormatMsg(char* output, uint8_t length, const char* format, unsigned int x1){
	uint8_t i, j=0;
	char temp[8];
	for(i=0; i<128; i++){
		if(j >= length) break;
		output[j] = format[i];
		if(format[i] == '\0') break;
		else if(format[i] == '%'){
			utoa(x1, temp, 10);
			uint8_t k=0;
			while(temp[k] != '\0'){
				output[j++] = temp[k++];
				if(j >= length) break;
			}
			i++;
		}else{
			j++;
		}
	}
	while(j<length) output[j++] = ' ';
	output[j] = '\0';
}

#define LOG_LINE_LENGTH 42
// Put 1 line of format string for logfile
void FormatLogFile(char* output){
	// format: "yyyy/mm/dd hh:mm:ss,[ch1],[ch2],[ch3],[ch4]\n"
	// buffer length: > 42  bytes
	// time
	uint8_t i;
	for(i=LOG_LINE_LENGTH-24; i<LOG_LINE_LENGTH-2; i++){
		output[i] = ' ';
	}
	uint8_t cnt = 2;
	output[0] = '2';
	output[1] = '0';
	BCD_print(&output[cnt], g_curTime.year); cnt += 2;
	output[cnt++] = '/';
	BCD_print(&output[cnt], g_curTime.month); cnt += 2;
	output[cnt++] = '/';
	BCD_print(&output[cnt], g_curTime.day); cnt += 2;
	output[cnt++] = ',';
	BCD_print(&output[cnt], g_curTime.hour); cnt += 2;
	output[cnt++] = ':';
	BCD_print(&output[cnt], g_curTime.min); cnt += 2;
	output[cnt++] = ':';
	BCD_print(&output[cnt], g_curTime.sec); cnt += 3;
	
	// value
	for(i=0; i<4; i++){
		output[cnt++] = ',';
		utoa(g_analogState[i], &output[cnt], 10);
		cnt += 4;
	}
	
	output[cnt++] = '\r';
	output[cnt] = '\n';
	return ;
}

#if (ENABLE_BRIGHTNESS)
static void UpdateBrightness(void){
	cli();
	UART_Send('b');
	UART_Send('A'+ g_eep.brightness);
	sei();
}
#endif

// Update Display (called every g_curDisplayState changes)
static void UpdateDisplay(void){
	cli();
	UpdateDisplayDigits();
	SendToDisplay((const uint8_t*)g_digitData);
	sei();
}


int main(void){
	
	Init();
	
	// Restore setting
#if (ENABLE_EEP)
	LoadEEP();
#else
	g_eep.brightness = 10;
#endif
#if (ENABLE_BRIGHTNESS)
	UpdateBrightness();
#endif
	
//	g_curDisplayState = STATE_THRU;
	// Update Display
	UpdateDisplay();
	
	static FATFS fs;
	UINT bw;
	
	uint8_t buf[10];
	
	while(1){
		uint8_t i;
		
		if(g_isrFlag){
		
			if(g_isrFlag & ISR_TIMER1_CALLED){
				// STOP TIMER1
				TCCR1B &= ~0x07;
				if(g_curDisplayState == STATE_CLOCK1){
					// clear colon
					g_digitData[8] &= 0b11110001;
					SendToDisplay((const uint8_t*)g_digitData);
				}
			}
			if(g_isrFlag & ISR_CHECK_SWITCH){
				// Check Switch State
				uint8_t sw = 0xFF;
				{
					uint8_t a = (g_analogState[4] >> 2);
					if(a < 0+32) sw = 0b11111110; // SW1: MENU
					else if(a < 64+32)  sw = 0b11111101; // SW2
					else if(a < 128+32) sw = 0b11111011; // SW3
					else if(a < 192+32) sw = 0b11110111; // SW4
				}
				
				GetSwitchState(sw, (uint8_t*)&g_swState);
				
				// Start acquiring the analog value
				ADMUX = (ADMUX & 0xF0) | 0;
				ADCSRA |= 0b01000000; // Start A-D conversion
			}
			
			if(g_isrFlag & ISR_PCINT1_CALLED){
				if(PINC & (1<<PC4)){
					// interrupt per 1sec.
					MSM62X42B_GetCurTime(&g_curTime);
					if(g_curDisplayState == STATE_CLOCK1){
						UpdateDisplayDigits();
						SendToDisplay((const uint8_t*)g_digitData);
					}
					
					// start 1-shot timer for colon blinking
					TCCR1B |= 0x05; // START TIMER1
					TCNT1   = 0;
					
					if(g_loggerStatus.logging){
						static uint16_t cnt2;
						if(cnt2++ >= g_loggerStatus.intervalCnt){ // Logging interval
							char buf[LOG_LINE_LENGTH];
							FormatLogFile(buf);
							pf_write(buf, LOG_LINE_LENGTH, &bw);
							if(bw < LOG_LINE_LENGTH){// file reached to the end
								pf_write(0, 0, &bw); // finalize
								g_loggerStatus.logging = 0;
							}
							cnt2 = 0;
						}
					}
				}
				
			}
			g_isrFlag = 0;
		}
		
		// check switch
		if(g_swState){
			if(g_swState == SW_SW1){ // MENU
				if(g_loggerStatus.logging){
					g_curDisplayState = STATE_LOGGER_SETTING_1;
					g_loggerStatus.logging = 0;
					pf_write(0, 0, &bw); // finalize
					strcpy_P( (char*)buf, PSTR("LOG END  "));
					buf[9] = 0xFF;
					SendToDisplay(buf);
					for(i=0; i<100; i++) _delay_ms(10);
					g_curDisplayState = STATE_CLOCK1;
					
				}
				else if(g_curDisplayState >= STATE_ADC1 && g_curDisplayState <= STATE_ADC4){
					g_curDisplayState = STATE_LOGGER;
#if (ENABLE_EEP)
					SaveEEP();
#endif
				}
				else if(g_curDisplayState <= STATE_THRU){
					g_curDisplayState ++;
					if(g_curDisplayState > STATE_THRU) g_curDisplayState = STATE_CLOCK1;
#if (ENABLE_EEP)
					SaveEEP();
#endif
				}else{
					g_curDisplayState = STATE_CLOCK1;
					MSM62X42B_EnableOutput();
				}
				
			}
			else if(g_swState == (SW_SW2 | SW_LONG)){ // OK + LONGPUSH
				if(g_curDisplayState == STATE_CLOCK1){ // Adjust time
					g_curDisplayState = STATE_ADJUST_TIME_YEAR;
					
					// Start time adjust and disable interrupt per 1 sec
					MSM62X42B_DisableOutput();
				}
			}
			else if(g_swState == (SW_SW2 )){ // OK
				if(g_curDisplayState >= STATE_ADJUST_TIME_YEAR && g_curDisplayState <= STATE_ADJUST_TIME_MIN){
					// Adjust time
					g_curDisplayState ++;
					if(g_curDisplayState > STATE_ADJUST_TIME_MIN) {
						// finish time adjust
						g_curTime.sec = 0;
						MSM62X42B_Reset();
						MSM62X42B_SetCurTime(&g_curTime);
						MSM62X42B_EnableOutput();
						g_curDisplayState = STATE_CLOCK1;
					}
				}
				else if(g_curDisplayState == STATE_LOGGER){ // Logger setting start
					// Try memory card initialization
					FRESULT ret;
					
					if((ret = pf_mount(&fs)) == FR_OK){
						// open file
						if((ret = pf_open(LOG_FILENAME)) == FR_OK){
							// pf_lseek(0);
							g_curDisplayState = STATE_LOGGER_SETTING_1;
						}
						else{
							// mount error
							strcpy_P( (char*)buf, PSTR("NO FILE "));
							buf[7] = '0'+ ret;
							buf[9] = 0xFF;
							SendToDisplay(buf);
							for(i=0; i<100; i++) _delay_ms(10); 
						}
					}else{
						// mount error
						strcpy_P( (char*)buf, PSTR("CARD ERR"));
						buf[7] = '0'+ ret;
						buf[9] = 0xFF;
						SendToDisplay(buf);
						for(i=0; i<100; i++) _delay_ms(10); 
					}
				}
				else if(g_curDisplayState == STATE_LOGGER_SETTING_1){
					SaveEEP();
					// mount error
					strcpy_P( (char*)buf, PSTR("LG START"));
					buf[9] = 0xFF;
					SendToDisplay(buf);
					for(i=0; i<100; i++) _delay_ms(10); 
					g_curDisplayState = STATE_ADC1;
					if(g_loggerStatus.intervalIdx == 1){
						g_loggerStatus.intervalCnt = 2-1;
					}
					else if(g_loggerStatus.intervalIdx == 2){
						g_loggerStatus.intervalCnt = 10-1;
					}
					else if(g_loggerStatus.intervalIdx == 3){
						g_loggerStatus.intervalCnt = 30-1;
					}
					else if(g_loggerStatus.intervalIdx == 4){
						g_loggerStatus.intervalCnt = 60-1;
					}
					else if(g_loggerStatus.intervalIdx == 5){
						g_loggerStatus.intervalCnt = 300-1;
					}
					g_loggerStatus.logging = 1;
				}
			}
			else if( (g_swState&0x0F) > SW_SW2 )
			{ // + or -
				if(g_curDisplayState >= STATE_ADC1 && g_curDisplayState <= STATE_ADC4){
					// select adc channel
					if((g_swState & 0x0F) == SW_SW3){
						g_curDisplayState ++;
						if(g_curDisplayState > STATE_ADC4) g_curDisplayState = STATE_ADC1;
					}else{
						g_curDisplayState --;
						if(g_curDisplayState < STATE_ADC1) g_curDisplayState = STATE_ADC4;
					}
				}
				else if(g_curDisplayState >= STATE_ADJUST_TIME_YEAR && 
					g_curDisplayState <= STATE_ADJUST_TIME_MIN){
					/*
					const uint8_t maxl[] = {0x99, 0x12, 0x31, 0x23, 0x59};
					const uint8_t minl[] = {0,1,1,0,0};
					uint8_t* pvall[] = {&(g_curTime.year), &(g_curTime.month),
						&(g_curTime.day), &(g_curTime.hour), &(g_curTime.min)};
					// Adjust time
					uint8_t  max  = maxl[g_curDisplayState-STATE_ADJUST_TIME_YEAR];
					uint8_t  min  = minl[g_curDisplayState-STATE_ADJUST_TIME_YEAR];
					uint8_t* pVal = pvall[g_curDisplayState-STATE_ADJUST_TIME_YEAR];
					*/
					
					
					uint8_t  max;
					uint8_t  min;
					uint8_t* pVal;
					if(g_curDisplayState == STATE_ADJUST_TIME_YEAR){
						max = 0x99;
						min = 0;
						pVal = &(g_curTime.year);
					}
					else if(g_curDisplayState == STATE_ADJUST_TIME_MONTH){
						max = 0x12;
						min = 1;
						pVal = &(g_curTime.month);
					}
					else if(g_curDisplayState == STATE_ADJUST_TIME_DAY){
						max = 0x31;
						min = 1;
						pVal = &(g_curTime.day);
					}
					else if(g_curDisplayState == STATE_ADJUST_TIME_HOUR){
						max = 0x23;
						min = 0;
						pVal = &(g_curTime.hour);
					}
					else{
						max = 0x59;
						min = 0;
						pVal = &(g_curTime.min);
					}
					
					if((g_swState & 0x0F) == SW_SW3){
						BCD_add(pVal);
						if(*pVal > max) *pVal = min;
					}else{
						BCD_dec(pVal);
						if((int8_t)(*pVal) < min) *pVal = max;
					}
				}
				else if(g_curDisplayState == STATE_LOGGER_SETTING_1){
					if((g_swState & 0x0F) == SW_SW3){
						g_loggerStatus.intervalIdx --;
						if(g_loggerStatus.intervalIdx > INTERVAL_MAX) g_loggerStatus.intervalIdx = INTERVAL_MAX;
					}else{
						g_loggerStatus.intervalIdx ++;
						if(g_loggerStatus.intervalIdx > INTERVAL_MAX) g_loggerStatus.intervalIdx = 0;
					}
				}
#if (ENABLE_BRIGHTNESS)
				else if(g_curDisplayState == STATE_CLOCK1){
					// adjust brightness
					if((g_swState & 0x0F) == SW_SW3){
						if(g_eep.brightness>2) g_eep.brightness --;
					}else{
						if(g_eep.brightness<36) g_eep.brightness ++;
					}
					UpdateBrightness();
#if (ENABLE_EEP)
					SaveEEP();
#endif
				}
#endif
			}
			// Update Display
			UpdateDisplay();
			g_swState = 0;
		}
	}
	
	return 0;
}
